Install

Example:

```
$ dart main.dart -o org.norc.dept -p myapp
```

This will generate the component bits in subdirectory `myapp`.  Test it:

```
$ cd myapp/myappDemo
$ flutter pub get
$ flutter run -d chrome
or
$ flutter run -d <device> # if you have a mobile connected
```

== Plugin Registration

Files for Android and iOS are generated even if no plugins are listed. This makes sense since the UI bit is cross-platform enabled by default.

Generated files:

* Android: `greetingsDemo/android/app/src/main/java/io/flutter/plugins/GeneratedPluginRegistrant.java`
* iOS
** `greetingsDemo/ios/Flutter/Generated.xcconfig`
** `greetingsDemo/ios/Runner/GeneratedPluginRegistrant.h`
** `greetingsDemo/ios/Runner/GeneratedPluginRegistrant.m`


==== Java package names.

The app has one package name; each plugin also has its own package
name. It doesn't matter what they are, but they must be different (?).

Package names go in several places:

* Each AndroidManifest.xml file (debug, main, profile)
* App:  `android/app/build.gradle`, in the `defaultConfig` stanza, key `applicationId`
* Plugin:  `android/build.gradle`, key `group`
** The Kotlin/Java implementation code, obviously;
** pubspec.yaml:
```
flutter:
  plugin:
    platforms:
      android:
        package: org.example.greetings  # Java package
        pluginClass: GreetingsPlugin
```



Bulding the app involves building for Android and iOS.  The build
logic always generates code to register plugins for the app, even if
it does not use any.

Plugins register themselves; that is, building them generates
registration code for each plugin.

If you screw up the plugin code your app might still build and run,
but the plugin functionality will be missing.

== Dependency Structure

* App lists the plugin as a dep:

.greetingsDemo/pubspec.yaml
[source,yaml]
----
dependencies:
  flutter:
    sdk: flutter
  greetings:   # this is the plugin
    path: plugins/greetings # this can be any path
----

* The plugin depends on the "adapter"

.greetingsDemo/plugins/greetings/pubspec.yaml
----
dependencies:
  flutter:
    sdk: flutter
  adapter:
    path: adapter
----

The plugin code imports the adapter code and exposes functions,
methods, etc. that wrap adapter.

* The adapter functions as an intermediary between client code (via
  the plugin) and the platform implementations.  It must implement
  `PlatformInterface` (defined by the plugin
  `plugin_platform_interface`), and it depends on each platform
  implementation:

.greetingsDemo/plugins/greetings/adapter/pubspec.yaml
----
dependencies:
  flutter:
    sdk: flutter
  plugin_platform_interface: ^1.0.2
  android:
    path: ../platforms/android
  ios:
    path: ../platforms/ios
  macos:
    path: ../platforms/macos
  web:
    path: ../platforms/web
----

NOTE: The adapter code itself does not import or refer to any of the
implementation code.  Rather, it's the other way around: each
implementation must depend on and import the adapter.  The dependency
is there in order to make the implementation code accessible by users
of the adapter.

NOTE: An app depending on the web implementation of the plugin will
include a file generated by `flutter pub get` (described below), which
imports and registers the web implementation package.  Listing the
latter as a dependency of the adapter is what makes it available to
the app.  IOW, the app does not explicitly list the plugin
implementations as deps; it only depends on the plugin, which depends
on the adapter, which depends on the implementations.

* Platform implementation dependencies depend on the type of platform.
  There are two platform types, "proxied" (those that use a
  `MethodChannel` for communication between client and implementation)
  and "direct" (those whose implementation is expressed in the dart
  language).  Android and iOS implementations are proxied; web
  implementations are not.  The latter must depend on the adapter:

.greetingsDemo/plugins/greetings/platforms/web/pubspec.yaml
----
dependencies:
  flutter:
    sdk: flutter
  adapter:
    path: ../../adapter
----

This is because the Android and iOS implementations do not include
dart code; they only include code written in a platform language
(Kotlin/Java for Android, Swift/ObjectiveC for iOS), which talks to
the dart client via MethodChannels.  The implementations are
registered as plugins by code that the flutter build tools generate.
The client-side communication via MethodChannel is handled by the
adapter code (which is wrapped by the plugin).

A web implementation, by contrast, is written in Dart and does not use
method channels; the implementation is registered as a plugin by the
generated code of the client app.

=== Notes

Running `flutter pub get` does more than resolve dependencies.  In the
case of apps using a web plugin, it also generates a file,
`generated_plugin_registrant.dart`, in the `lib` dir of the package
using the plugin.

This is controlled by the plugin's "adapter" package, which lists the
platform dependencies of the plugin.  If if depends on a web
implementation, then the registration file will be generated.
